(*****************************************************************
 ** MergeHull.sml
 **
 ** Copyright 2005 Umut A. Acar
 **
 ** Written by Jorge Vittes
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: A self-adjusting merge-hull implementation.
 **********************************************************************)

structure MergeHullAdaptive =
struct

structure Geo = GeometryTwoD
structure Point = Geo.Point
structure Poly = Polynomial
structure List = ListAdaptive

open Adaptive 
open List

exception Impossible

type pt = Point.t
type u = pt List.u
type t = pt List.t
type it = t
type ot = t

val minX = fn (a,b) => Poly.eval0sgn (Geo.minX (a, b))
val maxX = fn (a,b) => Poly.eval0sgn (Geo.maxX (a, b))
val orient = fn (a,(b,c)) => Poly.eval0sgn (Geo.lineSideTest ((b, c), a))

structure Hash = HashFamily(val M = 2)

fun splitTest () =
  let
    val hashFun = Hash.hashFun ()
    val test = fn pt => (hashFun o Word.toInt o Point.hash) pt = 0
  in
    test
  end


(* one pass *)
local
afun snoc (l, x) =
    case get $ l of NIL => put $ (CONS (x, put $ NIL))
      | CONS (h, t) => put $ (CONS (h, snoc $ (t, x)))

afun allExceptI (leftmost,l,p) =
    case l
     of NIL => put $ NIL
      | CONS (h,mt) =>
        case get $ mt
         of NIL => if Point.eq (get $ leftmost, p) orelse Point.eq (p,h)
                   then put $ NIL
                   else put $ (CONS (h, put $ NIL))
          | t => put $ (CONS (h, allExceptI $ (leftmost, t, p)))

val merge = afn (leftmost, h1, t1, h2, t2) =>
    let datatype knot =
                 T of {merge1: knot -> pt * u * pt * u -$> t,
                       merge2: knot -> pt * pt * (u * u) Box.t -$> t,
                       merge31: knot -> pt * t * pt * u -$> t,
                       merge32: knot -> pt * u * pt * t -$> t}
        val {putM, ...} = PutMemoTable.mkPut $ ()
        (* mergeHM' *)
        (* pre: l1 = l1' @ [p1], l2 = nil or l2 = l2' @ [p2] *)
        val merge1 = fn knot as T{merge2,...} => afn (p1, l1, p2, l2) =>
            let val ml12 = putM $ ((p1,p2), (l1, l2))
            in merge2 knot $ (p1, p2, ml12)
            end
        val merge2 = fn knot as T{merge1,merge31,merge32,...} =>
            memo (afn (p1, p2, ml12) =>
            let val (l1,l2) = get $ ml12
            in case l2
                of NIL =>
                   let 
                   in case l1
                       (* what about p2? *)
                       of NIL => put $ NIL (* ml1 or ml2 *)
                        | CONS (h1, mt1) =>
                          if orient (p2, (p1, h1))
                          then merge32 knot $ (p2, l2, h1, mt1)
                          else merge31 knot $ (h1, mt1, p2, l2)
                   end
                 | CONS (h2, mt2) =>
                   if orient (h2, (p1, p2))
                   then merge1 knot $ (p1, l1, h2, get $ mt2)
                   else case l1
                         of NIL => put $ (CONS (p2, allExceptI $ (leftmost, l2, p2)))
                          | CONS (h1, mt1) =>
                            if orient (h1, (p1, p2))
                            then merge31 knot $ (h1, mt1, p2, l2)
                            else merge32 knot $ (p2, l2, h1, mt1)
            end)
        (* addToListHA *)
        val merge31 = fn knot as T{merge1,...} => afn (p1, ml1, p2, l2) =>
            case get $ ml1
             of l1 as NIL =>
                if Point.eq (get $ leftmost, p1)
                then put $ NIL
                else put $ (CONS (p1, merge1 knot $ (p1, l1, p2, l2)))
              | l1 => put $ (CONS (p1, merge1 knot $ (p1, l1, p2, l2)))
        (* addToListB *)
        val merge32 = fn knot as T{merge1,...} => afn (p1, l1, p2, ml2) =>
            case l1
             of NIL =>
                if Point.eq (get $ leftmost, p1)
                then put $ NIL
                else put $ (CONS (p1, merge1 knot $ (p1, l1, p2, get $ ml2)))
              | _ => put $ (CONS (p1, merge1 knot $ (p1, l1, p2, get $ ml2)))
        val knot = T{merge1=merge1,merge2=merge2,merge31=merge31,merge32=merge32}
    in merge1 knot $ (h1, t1, h2, t2)
    end

(* h1 < h2 *)
val collate = afn (h1,mt1,t1,h2,mt2,t2) =>
    let val mt1 = snoc $ (mt1, h1)
        val mt2 = case t2
                   of NIL => put $ NIL
                    | _ => snoc $ (mt2, h2)
        val leftmost = put $ h1
    in put $ (CONS (h1, merge $ (leftmost, h1, get $ mt1, h2, get $ mt2)))
    end

(* mergeHM *)
val merge = afn (ml1, ml2) =>
    case (get $ ml1, get $ ml2)
     of (NIL, _) => ml2
      | (_, NIL) => ml1
      | (CONS (h1, mt1), CONS (h2, mt2)) =>
        case (get $ mt1, get $ mt2)
         of (NIL, NIL) =>
            if minX (h1, h2)
            then put $ (CONS (h1, put $ (CONS (h2, put $ NIL))))
            else put $ (CONS (h2, put $ (CONS (h1, put $ NIL))))
          | (t1, t2) =>
            if minX (h1, h2)
            then collate $ (h1, mt1, t1, h2, mt2, t2)
            else collate $ (h2, mt2, t2, h1, mt1, t1)

afun mergeHull l =
    if get $ (lengthLessThan 2 $ l)
    then l
    else let val (l1, l2) = partition (splitTest ()) $ l
             val l1 = mergeHull $ l1
             val l2 = mergeHull $ l2
             val l = merge $ (l1, l2)
         in l
         end

in
  val run = fn l => Adaptive.call (mergeHull, l)
end


(* two pass *)
local
(* mergeHM' *)
val merge = fn left => afn (p1,l1,p2,l2) =>
    let datatype knot =
                 T of {merge1: knot -> pt * u * pt * u -$> t,
                       merge2: knot -> pt * pt * (u * u) Box.t -$> t}
        val {putM, ...} = PutMemoTable.mkPut $ ()
        val merge1 = fn knot as T{merge2,...} => afn (p1,l1,p2,l2) =>
            let val ml12 = putM $ ((p1, p2), (l1, l2))
            in merge2 knot $ (p1, p2, ml12)
            end
        val merge2 = fn knot as T{merge1,...} =>
            memo (afn (p1, p2, ml12) =>
            let val (l1,l2) = get $ ml12
            in case l2
                of NIL =>
                   let in
                       case l1
                        of NIL =>
                           if left (p2,p1)
                           then put $ NIL
                           else put $ (CONS (p2, put $ NIL))
                         | CONS (h1,mt1) =>
                           if orient (p2, (p1,h1))
                           then put $ (CONS (p2, merge1 knot $ (p2, l2, h1, get $ mt1)))
                           else if left (h1, p2)
                           then put $ (CONS (h1, merge1 knot $ (h1, get $ mt1, p2, l2)))
                           else put $ l1
                   end
                 | CONS (h2,mt2) =>
                   if orient (p1, (p2, h2))
                   then merge1 knot $ (p1, l1, h2, get $ mt2)
                   else case l1
                         of NIL => put $ (CONS (p2, put $ l2))
                          | CONS (h1, mt1) =>
                            if orient (p2, (p1, h1))
                            then put $ (CONS (p2, merge1 knot $ (p2, l2, h1, get $ mt1)))
                            else put $ (CONS (h1, merge1 knot $ (h1, get $ mt1, p2, l2)))
            end)
        val knot = T{merge1=merge1, merge2=merge2}
    in merge1 knot $ (p1,l1,p2,l2)
    end

(* mergeHM *)
val merge = fn left => afn (ml1,ml2) =>
    case (get $ ml1, get $ ml2)
     of (NIL, _) => ml2
      | (_, NIL) => ml1
      | (CONS (h1,mt1), CONS (h2,mt2)) =>
        let val (t1, t2) = (get $ mt1, get $ mt2)
        in if left (h1,h2)
           then put $ (CONS (h1, merge left $ (h1, t1, h2, t2)))
           else put $ (CONS (h2, merge left $ (h2, t2, h1, t1)))
        end

fun halfHull left = afn ml =>
    if get $ (lengthLessThan 2 $ ml)
    then ml
    else let val (ml1, ml2) = partition (splitTest ()) $ ml
             val ml1 = halfHull left $ ml1
             val ml2 = halfHull left $ ml2
             val ml = merge left $ (ml1, ml2)
         in ml
         end

afun dropLast ml =
   case get $ ml
    of NIL => put $ NIL
     | CONS (h,mt) =>
       case get $ mt
        of NIL => put $ NIL
         | CONS _ => put $ (CONS (h, dropLast $ mt))

val dropFirstLast = afn ml =>
    case get $ ml
     of NIL => put $ NIL
      | CONS (h, mt) =>
        dropLast $ mt

val mergeHull = afn ml =>
    let val lhull = halfHull maxX $ ml
        val uhull = halfHull minX $ ml
        val lhull = dropFirstLast $ lhull
(*
            case get $ lhull
             of NIL => put $ NIL
              | CONS (h, mt) => removeEnd $ mt
*)
        val ml = append $ (uhull, lhull)
    in ml
    end

in
val run = fn ml => Adaptive.call (mergeHull, ml)
end

end
