(**********************************************************************
 ** hash-table.sml
 **
 ** Copyright 2008 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: The implementation of hash tables
 **********************************************************************)

structure PolyHashTable :> POLY_HASH_TABLE =
struct

   val initSize = 4 (* must be a power of 2 *)
   val sizeToMask = fn size => Word.fromInt size - 0w1
   val growthRatio = 8
   val bucketRatio = 2

   type ('key,'value) bucket = {hash: word, key: 'key, value: 'value} list
   datatype ('key,'value) table = 
      T of {bucketsR: ('key,'value) bucket array ref,
            eqF: 'key Eq.t,
            hashF: 'key Hash.t,
            maskR: word ref,
            numItemsR: int ref}
   type ('key,'value) t = ('key,'value) table
      
   val new: 'key Key.t -> ('key,'value) t =
      fn {eq, hash} =>
      T {bucketsR = ref (Array.array (initSize, [])),
         eqF = eq,
         hashF = hash, 
         maskR = ref (sizeToMask initSize),
         numItemsR = ref 0}

   fun pushOntoBucket (buckets: ('k,'v) bucket array, i: int, hkv) =
      Array.update (buckets, i, hkv :: (Array.sub (buckets, i)))

   fun growTable (table: ('k,'v) t, newSize: int) : unit =
      let
         val T {bucketsR, maskR, ...} = table
         val buckets = !bucketsR
         val newBuckets = Array.array (newSize, [])
         val newMask = sizeToMask newSize
         fun copyHKV (hkv as {hash, ...}) =
            let
               val i = Word.toIntX (Word.andb (hash, newMask))
               val () = pushOntoBucket (newBuckets, i, hkv)
            in
               ()
            end
         val () = Array.app (List.app copyHKV) buckets
         val () = bucketsR := newBuckets
         val () = maskR := newMask
      in
         ()
      end

   fun insert (table: ('k,'v) t, key: 'k, value: 'v) : unit =
      let
         val T {bucketsR,hashF,maskR,numItemsR,...} = table
         val buckets = !bucketsR
         val mask = !maskR
         val numItems = !numItemsR
         val hkv = {hash = hashF key, key = key, value = value}
         val i = Word.toIntX (Word.andb (#hash hkv, mask))
         val () = pushOntoBucket (buckets, i, hkv)
         val () = numItemsR := numItems + 1
         val size = Array.length buckets
         val () = if numItems >= bucketRatio * size
                     then growTable (table, growthRatio * size)
                  else ()
      in
         ()
      end

   fun find (table: ('k,'v) t, key': 'k, valueP': 'v -> bool) : 'v option =
      let
         val T {bucketsR,eqF,hashF,maskR,...} = table
         val buckets = !bucketsR
         val mask = !maskR
         val hash' = hashF key'
         val i = Word.toIntX (Word.andb (hash', mask))
         val hkvOpt =
            List.find (fn {hash,key,value} =>
                       hash' = hash 
                       andalso eqF (key', key) 
                       andalso valueP' value)
                      (Array.sub (buckets, i))
         val vOpt =
            Option.map (fn {value,...} => value) hkvOpt
      in
         vOpt
      end

   fun remove (table: ('k,'v) t, key': 'k, valueP': 'v -> bool) : unit =
      let
         val T {bucketsR,eqF,hashF,maskR,numItemsR,...} = table
         val buckets = !bucketsR
         val mask = !maskR
         val numItems = !numItemsR
         val hash' = hashF key'
         val i = Word.toIntX (Word.andb (hash', mask))
         fun loop l =
            case l of
               [] => (false, [])
             | (hkv as {hash,key,value}) :: l =>
                  if hash' = hash 
                     andalso eqF (key', key) 
                     andalso valueP' value
                     then (true, l)
                  else let val (b,l) = loop l
                       in (b, hkv :: l)
                       end 
         val (b, l) = loop (Array.sub (buckets, i))
         val () =
            if b
               then (Array.update (buckets, i, l)
                     ; numItemsR := numItems - 1)
            else raise Fail "HashTable.remove"
      in
         ()
      end
end

(*
functor BuildHashTable(structure Key: HASH_TABLE_KEY) :> 
   MONO_HASH_TABLE where type Key.t = Key.t =
struct
   structure Key = Key
   open PolyHashTable

   type 'value t = (Key.t, 'value) t
   val new = fn () => new {eq = Key.eq, hash = Key.hash}
end
*)