(**********************************************************************
 ** self-adjust-core.sml
 **
 ** Copyright 2008 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: The implementation of the core self-adjusting code
 **********************************************************************)

structure SelfAdjustCore :> SELF_ADJUST_CORE =
struct 

type 'a cont = 'a -> unit

(**********************************************************************
 ** Debugging
 **********************************************************************)
val DEBUG = false

fun dPrint f = 
   if DEBUG then
      print (f ())
   else
      ()

(**********************************************************************
 ** Reuse flag
 **********************************************************************)
structure ReuseFlag :> 
sig  
   val get: unit -> bool
   val setTrue: unit -> unit
   val setFalse: unit -> unit
end =
struct
   val RF : bool ref = ref false

   val get = fn () => !RF
   val setTrue = fn () => RF := true
   val setFalse = fn () => RF := false
end

(**********************************************************************
 ** Time Stamps
 **********************************************************************)
structure TimeStamps =
struct
   open TimeStamps

   val START = ref TimeStamps.special
   val STOP = ref TimeStamps.special
   val NOW = ref TimeStamps.special
      
   val reset = fn () =>
      let 
         val ts = TimeStamps.init ()
         val () = START := ts
         val () = STOP := TimeStamps.add ts
         val () = NOW := TimeStamps.add ts
      in 
         ()
      end
   
   fun insert () = 
      let 
         val t = TimeStamps.add (!NOW) 
         val () = NOW := t
      in
         t
      end

   fun stats () =
      TimeStamps.size (!START)
end

(**********************************************************************
 ** PropagateQueueElt
 **********************************************************************)
structure PropagateQueueElt = 
struct
   type t = TimeStamps.t * (unit -> unit)
   fun cmp ((ta,_), (tb,_)) = TimeStamps.compare(ta,tb)
end

(**********************************************************************
 ** PropagateQueue
 **********************************************************************)
structure PropagateQueue :
sig
   val reset: unit -> unit
   val enqRead: PropagateQueueElt.t -> unit
   val enqWrite: PropagateQueueElt.t -> unit
   val deq: unit -> PropagateQueueElt.t option
end =
struct
   structure H = BuildLeftistHeap (structure Elt = PropagateQueueElt)

   val PQ = ref H.empty
      
   val reset = fn () => PQ := H.empty
      
   val enq = fn e => PQ := H.insert (!PQ, e)

   val enqRead = fn (e as (t, _)) =>
      let
         val _ = dPrint (fn () => 
                         "--> Queueing read start = " 
                         ^ TimeStamps.toString t ^ "\n")
      in
         enq e
      end
   
   val enqWrite = fn (e as (t, _)) =>
      let
         val _ = dPrint (fn () => 
                         "--> Queueing write start = " 
                         ^ TimeStamps.toString t ^ "\n")
      in
         enq e
      end
   
   fun deq () =
      case H.deleteMin (!PQ) of
         NONE => NONE
       | SOME (q,e as (t,_)) =>
            (PQ := q
             ; if TimeStamps.isSplicedOut t
                  then deq ()
               else SOME e)
end

(**********************************************************************
 ** Readers
 **********************************************************************)
structure Reader = PropagateQueueElt

(**********************************************************************
 ** Modifiables
 **********************************************************************)
structure Modref :>
sig
   datatype 'a t = T of word * 'a Eq.t * 'a option ref * Reader.t PolyCircularList.t

   val getEq : 'a t -> 'a Eq.t
   val getValOptPtr : 'a t -> 'a option ref
   val getValOpt : 'a t -> 'a option
   val getVal : 'a t -> 'a
   val getReaders : 'a t -> Reader.t PolyCircularList.t
   val hasNoReads : 'a t -> bool

   val newEmpty : 'a Eq.t -> 'a t
   val newWrite : 'a Eq.t * 'a -> 'a t

   val eq: 'a t Eq.t
   val hash: 'a t Hash.t
   val key: 'a t Key.t
   val toString : 'a t -> string
end =

struct
   datatype 'a t = T of word * 'a Eq.t * 'a option ref * Reader.t PolyCircularList.t

   val ID : word ref = ref 0wx0
   val nextID = fn () => 
      let
         val w = !ID
         val () = ID := w + 0w1
      in
         w
      end

   val getId = fn (T (w,_,_,_)) => w
   val getEq = fn (T (_,eq,_,_)) => eq
   val getValOptPtr = fn (T (_,_,vop,_)) => vop
   val getValOpt = fn modr => ! (getValOptPtr modr)
   val getVal = fn modr => valOf (getValOpt modr)
   val getReaders = fn (T (_,_,_,rs)) => rs
   val hasNoReads = fn (T (_,_,_,rs)) => PolyCircularList.isEmpty rs

   val new = fn (eq,vo) => T (nextID (), eq, ref vo, PolyCircularList.empty ())
   val newEmpty = fn eq => new (eq, NONE)
   val newWrite = fn (eq,x) => new (eq, SOME x)

   val eq = fn (m1,m2) => (getValOptPtr m1) = (getValOptPtr m2)
   val hash = fn m => getId m
   val key = {eq = eq, hash = hash}
   val toString = fn m => "< ModRef " ^ Word.toString (getId m) ^ " >"
end
type 'a modref = 'a Modref.t
structure Dest =
struct
   type 'a t = 'a modref
end
type 'a dest = 'a Dest.t

(** Assign **)
val assign = fn (modr, x) =>
   let
      val _ = dPrint (fn () => "writing modr = " ^  Modref.toString modr ^ "\n")
      val valueOptPtr = Modref.getValOptPtr modr
   in      	
      case !valueOptPtr of
         NONE => valueOptPtr := SOME x
       | SOME x' =>
            if (Modref.getEq modr) (x,x')
               then ()
            else (valueOptPtr := SOME x
                  ; PolyCircularList.foreach 
                    (Modref.getReaders modr, PropagateQueue.enqRead))
   end

(** Alloc **)
val alloc : ('a Eq.t * 'a dest cont) cont = fn (eq, cont) =>
   let 
      val modr = Modref.newEmpty eq
      val _ = dPrint (fn () => "allocating modr = " ^  Modref.toString modr ^ "\n")
   in 
      cont modr
   end

(** Write **)
val write : ('a dest * 'a * 'a modref cont) cont = fn (modr, x, cont) =>
   let 
      val _ = dPrint (fn () => "writing modr = " ^  Modref.toString modr ^ "\n")
      val () = assign (modr, x)
   in 
      cont modr
   end

(** Read **)
val read : ('a modref * 'a cont) cont = fn (modr, cont) =>
   let
      val _ = dPrint (fn () => "reading modr = " ^  Modref.toString modr ^ "\n")

      val readers = Modref.getReaders modr

      fun run () = let val x = Modref.getVal modr in cont x end

      val t = TimeStamps.insert ()
      val readerElt = PolyCircularList.insert (readers, (t, run))
      val _ = TimeStamps.setInv (t, fn () => PolyCircularList.remove (readers, readerElt))

      val () = run ()
   in
      ()
   end


(**********************************************************************
 ** Meta operations
 **********************************************************************)

(** Init **)
fun init () =
   let
      val () = TimeStamps.reset ()
      val () = PropagateQueue.reset ()
   in
      ()
   end

(** New **)
fun new (eq, x) = Modref.newWrite (eq, x)

(** Change **)
fun change (modr, x) = assign (modr, x)

(** Deref **)
fun deref modr =
  case Modref.getValOpt modr of
     NONE => raise Fail "deref"
   | SOME x => x

(** Propagate **)
fun propagate () =  
   let
      fun loop r = 
         if ReuseFlag.get () then 
            case PropagateQueue.deq () of
               NONE => (r, true)
             | SOME (t, f) =>
                  let
                     val _ = TimeStamps.NOW := t
                     val _ = ReuseFlag.setFalse ()
                     val _ = f ()
                  in
                     loop (r + 1)
                  end
         else
            (TimeStamps.spliceOut (!TimeStamps.NOW, !TimeStamps.STOP)
             ; (r, false))

      val _ = ReuseFlag.setTrue ()
      val res = loop 0
   in	    
      res
   end

val hasNoReads : 'a modref  -> bool = Modref.hasNoReads

val stats = fn () =>
   ("num timestamps: " ^ (Int.toString (TimeStamps.stats ())) ^ "\n")
   :: nil

(**********************************************************************
 ** Memoization
 **********************************************************************)
structure MemoTable =
struct
   val doSkip = fn time =>
      TimeStamps.isSplicedOut time
      orelse TimeStamps.compare (time, !TimeStamps.NOW) = LESS

    fun spliceIn t = 
       let
          val _ = TimeStamps.spliceOut (!TimeStamps.NOW,t)
          val _ = TimeStamps.NOW := t
          val _ = ReuseFlag.setTrue ()
       in
          ()
       end

    (** mkMemoFunOne, mkMemoFunTwo **)
    local
       val memoFun = 
          fn (keyArg : 'arg Key.t) =>
          fn {create: 'res cont * exn cont -> 'mkh,
              update: 'mkh * ('res cont * exn cont) -> unit} => 
          let 
             val table = PolyHashTable.new keyArg
          in
             fn (f: ('arg * 'mkh) cont) =>
             fn (arg: 'arg, k: 'res cont, h: exn cont) => 
             case PolyHashTable.find (table, arg, fn (t,_) => not (doSkip t)) of
                NONE =>
                   let
                      val t = TimeStamps.insert ()
                      val mkh = create (k,h)
                      val () = f (arg, mkh)
                      val () = PolyHashTable.insert (table, arg, (t, mkh))
                      val delete = fn () => 
                         PolyHashTable.remove 
                         (table, arg, fn (t',_) => 
                          TimeStamps.compare (t, t') = EQUAL)
                      val () = TimeStamps.addToInv (t, delete)
                   in
                      ()
                   end
              | SOME (t,mkh) =>
                   let
                      val () = update (mkh, (k,h))
                      val () = spliceIn t
                   in
                      ()
                   end
          end
    in
       (* use memo table for mkh *)
       val mkMemoFunOne = fn eq_hash =>
          let
             fun create (k,h) = 
                Modref.newWrite 
                (Eq.tuple2 (Eq.pointer, Eq.pointer), 
                 (k, h))
             fun update (mkh, (k,h)) = assign (mkh, (k,h))
          in
             memoFun eq_hash {create = create, update = update}
          end
       (* use memo table for (mk,mh) *)
       val mkMemoFunTwo = fn eq_hash =>
          let
             fun create (k, h) =
                (Modref.newWrite (Eq.pointer, k),
                 Modref.newWrite (Eq.pointer, h))
             fun update ((mk,mh),(k,h)) =
                (assign (mk, k)
                 ; assign (mh, h))
          in
             memoFun eq_hash {create = create, update = update}
          end
    end

    (* mkMemoFun *)
    local
       val memoFun = 
          fn (keyArg: 'arg Key.t) =>
          let 
             val table = PolyHashTable.new keyArg
          in
             fn (f: 'arg cont) =>
             fn (arg: 'arg) => 
             case PolyHashTable.find (table, arg, fn t => not (doSkip t)) of
                NONE =>
                   let
                      val t = TimeStamps.insert ()
                      val () = f arg
                      val () = PolyHashTable.insert (table, arg, t)
                      val delete = fn () => 
                         PolyHashTable.remove 
                         (table, arg, fn t' => 
                          TimeStamps.compare (t, t') = EQUAL)
                      val () = TimeStamps.addToInv (t, delete)
                   in
                      ()
                   end
              | SOME t =>
                   let
                      val () = spliceIn t
                   in
                      ()
                   end
          end
    in
       val mkMemoFun = memoFun
    end

    (* mkMemoCont *)
    local
       val memoCont = 
          fn (eq: 'arg Eq.t) => 
          fn (k: 'arg cont) =>
          let
             val table: (TimeStamps.t * 'arg) option ref = ref NONE
          in
             fn (arg: 'arg) =>
             let
                fun run () = 
                   let
                      val t = TimeStamps.insert ()
                      val _ = table := SOME (t, arg)
                      val delete = fn () =>
                         Option.app 
                         (fn (t',_) =>
                          if TimeStamps.compare (t, t') = EQUAL
                             then table := NONE
                          else ()) (!table)
                      val _ = TimeStamps.addToInv (t, delete)
                   in
                      k arg
                   end
             in
                case (!table) of
                   NONE => run ()
                 | SOME (t',arg') => 
                      if eq (arg,arg') then
                         spliceIn t'
                      else
                         run ()
             end
          end
    in
       val mkMemoCont = memoCont
    end
end

structure AllocWriteTable =
struct
   structure Dest =
      struct
         datatype 'a t = T of 'a Eq.t * TimeStamps.t
      end

   structure Table =
      struct
         type ('key, 'a) t =
            ('key, {allocTime: TimeStamps.t,
                    writeTime: TimeStamps.t,
                    modrOptRef: 'a Modref.t option ref,
                    cont: 'a Modref.t cont}) PolyHashTable.t
         val make : 'key Key.t -> ('key, 'a) t = 
            PolyHashTable.new
      end

   val writeEntry = 
      fn {table: ('key, 'a) Table.t,
          allocTime: TimeStamps.t,
          writeTime: TimeStamps.t,
          key: 'key,
          eq: 'a Eq.t,
          value: 'a,
          cont: 'a Modref.t cont} =>
      let
         val doSkip = fn {allocTime,modrOptRef,...} =>
            TimeStamps.isSplicedOut allocTime
            orelse TimeStamps.compare (allocTime, !TimeStamps.NOW) = LESS
            orelse not (isSome (!modrOptRef))
         val delete = fn () => 
            PolyHashTable.remove 
            (table, key, fn {writeTime = writeTime', ...} =>
             TimeStamps.compare (writeTime, writeTime') = EQUAL)
         val () = TimeStamps.addToInv (writeTime, delete)
      in
         case PolyHashTable.find (table, key, not o doSkip) of
            NONE => 
               let
                  val modr = Modref.newWrite (eq, value)
                  val () = 
                     PolyHashTable.insert
                     (table, key,
                      {allocTime = allocTime, writeTime = writeTime,
                       modrOptRef = ref (SOME modr), cont = cont})
               in
                  cont modr
               end
          | SOME {writeTime = writeTime',
                  modrOptRef = modrOptRef' as ref (modrOpt' as SOME (modr' as Modref.T (_, eq', ref (SOME value'), _))),
                  cont = cont', ...} => 
               let
                  (* steal *)
                  val () =
                     PolyHashTable.insert
                     (table, key,
                      {allocTime = allocTime, writeTime = writeTime, 
                       modrOptRef = ref modrOpt', cont = cont})
                  val () = modrOptRef' := NONE

                  fun rerunWrite () = 
                      let
                         val modr' = Modref.newWrite (eq', value')
                         val () = modrOptRef' := SOME modr'
                         val () = cont' modr'
                      in
                         ()
                      end

                  val _ = PropagateQueue.enqWrite (writeTime', rerunWrite)
                  val _ = assign (modr', value)
               in
                      cont modr'
               end
      end

   val alloc: ('a Eq.t * 'a Dest.t cont) cont = 
      fn (eq,cont) =>
      let 
         val allocTime = TimeStamps.insert ()
      in
         cont (Dest.T (eq,allocTime))
      end
   val write : ('key,'a) Table.t * 'key -> ('a Dest.t * 'a * 'a modref cont) cont =
      fn (table, key) => fn (dest,x,cont) =>
      let
         val Dest.T (eq,allocTime) = dest
         val writeTime = TimeStamps.insert ()
      in
         writeEntry {table = table,
                     allocTime = allocTime,
                     writeTime = writeTime,
                     key = key,
                     eq = eq,
                     value = x,
                     cont = cont}
      end

   val allocWrite : ('key,'a) Table.t * 'key -> ('a Eq.t * 'a * 'a modref cont) cont =
      fn (table,key) => fn (eq,x,cont) =>
      let
         val allocWriteTime = TimeStamps.insert ()
      in
         writeEntry {table = table,
                     allocTime = allocWriteTime,
                     writeTime = allocWriteTime,
                     key = key,
                     eq = eq,
                     value = x,
                     cont = cont}
      end
end

end 
