(**********************************************************************
 ** Test.sml
 **
 ** Copyright 2005 Umut A. Acar
 **
 ** Written by Kanat Tangwongsan and Jorge Vittes
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: The  polynomial root solver.
 **********************************************************************)
structure PSolver : POLYNOMIAL_SOLVER =
 struct
 structure M = Math
 structure R = Real

 open Polynomial       

 fun asinh x = M.ln(x + M.sqrt(1.0 + x*x))
 fun acosh x = M.ln(x + M.sqrt(x*x - 1.0))
	       
 val ieps : real = 1E~7

 (* Solve x^3 + p*x + q = 0 for all lists 
  *
  * Currently outputs only real roots that flip the sign in ASCENDING order i.e. no need to sort anymore.
  *)
 fun findAllRealVietaForm (p, q) =
     if (R.==(R.abs p, 0.0)) then 
	 [(Real.fromInt (Real.sign (~q)))*M.pow(Real.abs q,1.0/3.0)]
     else
	 let
	     val C = ~q/2.0*M.pow(3.0/(Real.abs p), 1.5)
	     val yl = 
		 if (R.>(p,0.0)) then [M.sinh ((asinh C)/3.0)]
		 else (if (R.>=(C,1.0)) then [M.cosh ((acosh C)/3.0)] 
		       else if (R.<= (C,~1.0)) then 
			   [~(M.cosh ((acosh (~C))/3.0))]
		       else 
			   (let val y0 = M.acos C
			    in  List.map (fn i => M.cos((y0 + (R.fromInt (2*i))*M.pi)/3.0)) [1, 2, 0]
			    end))
		      
	     val mult = 2.0*M.sqrt((R.abs p)/3.0)
	 in
	     List.map (fn x=>mult*x) yl
	 end


 fun findAllRealNormal (b,c,d) =
     let
	 val lambda = b/3.0
	 val p = (c - b*b/3.0)
	 val q = (d - b*c/3.0 + 2.0*b*b*b/27.0)
	 val rawroots = findAllRealVietaForm (p,q)
     in
	 List.map (fn r => r - lambda) (findAllRealVietaForm (p,q))
     end
     
 exception InfinitelyMany

 exception badCode

 fun constantSolver(p) = nil

 fun linearSolver(p:real list) = 
     case p of 
	 [b,a] => [(~b/a, true)]
        | _ => raise badCode

 fun quadraticSolver(p:real list) =
     case p of
	 [c,b,a] =>
            let
		val sq:real = b*b - 4.0*a*c
            in
                case Real.compare(sq,0.0) of
		    LESS => []
                  | EQUAL => [(~b/(2.0*a), false)]
		  | GREATER => 
		    let
			val sqr = M.sqrt(sq)
		        val twoa = 2.0*a
		    in
			[((~b-sqr)/twoa,true),(((~b+sqr)/twoa),true)]
                    end
	    end
	| _ => raise badCode

 fun cubicSolver(p) = 
     case p of
	[d,c,b,a] => List.map (fn x => (x,true)) (findAllRealNormal (b/a,c/a,d/a))
      | _ => raise badCode

 fun printList(p) =
     case p of
	 nil => ()
       | h::t => (print (" " ^ Real.toString(h) ^ ",");
                  printList(t))
                   



exception DegreeTooHigh

 (* a*x^3 + b*x^2 + c*x + d *)
 fun findAllReals (p,len) =
     case len of
	 0 => []
       | 1 => constantSolver(p)
       | 2 => linearSolver(p)
       | 3 => quadraticSolver(p)
       | 4 => cubicSolver(p)
       | _ => raise DegreeTooHigh



 fun cubicPolyInTuple l =
     let
	 fun padZero (n:int) (list:real list) =
	     case list of
		 h::t => h::(padZero ((n-1)) t)
	       | nil => 
		 if n>0 then 
		     0.0::(padZero ((n-1)) nil)
		 else nil
		      
	 val cubic = padZero 4 l
	 val d::c::b::a::t = cubic
     in
	 (a,b,c,d)
     end
 

fun sortl l =
  case l of
    nil => nil
  | h::t => 
      let 
	fun insert list (x as (xval,_)) =
	  case list of
	     nil => [x]
	   | (h as (x',_))::t => if (Real.<(xval, x')) then x::(h::t)
		     else h::(insert t x)
	val t' = sortl t
      in 
	insert t' h 
      end

 fun compareVals(h,h2) =
     let
	 val (v1,b1) = h
         val (v2,b2) = h2
         val b = (Real.==(v1,v2) andalso (b1 = b2))
         val _ = if not b then
		    (print ("\nv1 =" ^ Real.toString(v1) ^ "\n" ^
                            "v2 =" ^ Real.toString(v2)))
		 else ()
     in
	 b
     end

 fun compareLists(a,b) =
     case a of 
	 nil =>
	   (case b of
	      nil => true
            | _ => false)
       | h::t =>
	   case b of
	       h2::t2 => 
	          if not(compareVals(h,h2)) then
		      false
                  else
		      compareLists(t,t2)
             | _ => false


   fun getLengthAndTrim(p) =
    let
	fun trim(pol) =
	    case pol of 
		nil => (0,nil,false)
              | h::t =>
		 let val (len,p',v) = trim(t)
		 in if (not v) then 
		      if Real.==(h,0.0) then 
			(0,nil,false)
                     else
			(1,[h],true)
	            else
			(len+1,h::p',true)
	         end
       
        val (length,poly,_) = trim(p)

    in
	(poly,length)
    end


 fun allRealFlag (p) =
     let     
	  val (poly',length) = getLengthAndTrim p
     in
	 sortl (findAllReals (poly',length))
     end

     
 fun allReal (p) =
     List.map (#1) (allRealFlag p)

 fun allToggleSign (p) = 
     List.map (#1) (List.filter (#2) (allRealFlag p))

 end
