(**********************************************************************
 **
 ** Binary Heap Implementation
 ** 
 ** Umut Acar/Kanat Tangwongsan
 **
 **********************************************************************)
functor BinaryHeap (structure Priority : TOTAL_ORD_VALID) =
struct
type elt = Priority.t

type pqueue = {num_entries : int, 
	       cleanup_thres : int, 
	       pq : elt array}

type t = pqueue

val initSize = 16
exception Empty
exception InternalError_BinaryHeap

fun new () = ref {num_entries = 0, 
		  (* INVARIANT: cleanup_thres <= array size *)
		  cleanup_thres = initSize, 
		  pq = Array.array(initSize, Priority.EMPTY)}

fun reset heap = heap := !(new())

fun size heap =
  let val {num_entries, pq, cleanup_thres } = !heap 
  in num_entries
  end


fun countValid heap = 
  let
    val {num_entries, pq, cleanup_thres } = !heap 
					    
    fun loop (i, cnt) = 
      if (i >= num_entries) then cnt
      else 
	(case Priority.isValid (Array.sub (pq, i)) of
	   true => loop (i+1,cnt+1)
	 | _ => loop(i+1, cnt))
  in
    loop (0,0)
  end
fun isEmpty heap = (size heap) <= 0
      

fun isValid heap =
  let
    val {num_entries, pq, cleanup_thres} = !heap

    fun minLoc ((p1, v1), p2) =
      if (p2 < num_entries) then
	let 
	  val v2 = Array.sub(pq, p2)
	in
	  case Priority.compare(v2, v1) of
	    LESS => (p2, v2)
	  | _ => (p1,v1)
	end
      else (p1,v1)

    fun isValidAtLoc loc =
      let
	val myValue = Array.sub(pq, loc)
	val cl = minLoc ((loc,myValue), 2*loc + 1)
	val (p', v') = minLoc (cl, 2*loc + 2)
	val _ = if Priority.compare(myValue, Priority.EMPTY) = EQUAL then 
		  (print "valid check failed"; raise InternalError_BinaryHeap) 
		else ()
      in
	p' = loc
      end
      
    fun loop i =
      if (i >= num_entries) then true
      else (isValidAtLoc i) andalso (loop (i+1))
  in
    loop 0
  end

fun percolateDown (num_entries, pqArray, loc) = 
  let
    fun minLoc ((p1, v1), p2) =
      if (p2 < num_entries) then
	let 
	  val v2 = Array.sub(pqArray, p2)
	in
	  case Priority.compare(v2, v1) of
	    LESS => (p2, v2)
	  | _ => (p1,v1)
	end
      else (p1,v1)

    val myValue = Array.sub(pqArray, loc)
    val cl = minLoc ((loc,myValue), 2*loc + 1)
    val (p', v') = minLoc (cl, 2*loc + 2)
  in
    if (p' = loc) then
      ()
    else 
      let
	val _ = Array.update(pqArray, loc, v')
	val _ = Array.update(pqArray, p', myValue)		  
      in
	percolateDown (num_entries, pqArray, p')
      end
  end

fun percolateUp (myValue, pqArray, loc) = 
  let      
    val next = (loc - 1) div 2
  in
    if (next < 0) then
      Array.update(pqArray, loc, myValue)
    else
      let val nextValue = Array.sub(pqArray, next)
      in
	case Priority.compare(myValue, nextValue) of
	  LESS => (Array.update (pqArray, loc, nextValue);
		   percolateUp (myValue, pqArray, next))
	| _ => Array.update(pqArray, loc, myValue)
      end
  end


fun renewHeap (heap, justCopy) =
  let
    val {num_entries, cleanup_thres, pq} = !heap       
    val pq_new = Array.array(num_entries * 2 + 1, Priority.EMPTY)
    val oldSize = num_entries


    fun copyLive (i, j) =
      if (i >= oldSize) then j
      else
	let val data = Array.sub(pq, i)
	in
	  case justCopy orelse (Priority.isValid data) of
	    true => (Array.update(pq_new, j, data);
		     copyLive(i+1,j+1))
	  | false => copyLive(i+1,j)
	end

    val liveStock = copyLive (0,0)

    fun buildHeap () =
      let
	fun loop i =
	  if (i >= 0) then
	    (percolateDown (liveStock, pq_new, i);
	     loop (i-1))
	  else
	    ()
      in
	loop ((liveStock-1) div 2)
      end

    val _ = buildHeap ();

    val thres' = if justCopy then cleanup_thres else 3*liveStock

    val _ = if justCopy then ()
	    else print ("renewHeap: old="^(Int.toString num_entries)^
	                ",new="^(Int.toString liveStock)^
	                ",thr="^(Int.toString thres')^
	                ",len="^(Int.toString (Array.length pq_new))^"\n")
		 
  in
    (* Note that liveStock <= num_entries, 
     * so 2 * liveStock <= 2 *num_entries + 1 *)      
    heap := {num_entries = liveStock, cleanup_thres = thres',
	     pq = pq_new}
  end

fun insert (x, heap) = 
  let
    val {num_entries, cleanup_thres, pq} = !heap 
    val _ = 
	if (num_entries >= (Array.length pq)) then 
	  renewHeap (heap, true)  
        (** Cannot garbage collect the heap when
            inserting---hasNoReads may hold for non-garbage entries **)
	else ()

    val {num_entries, cleanup_thres, pq} = !heap 


    val _ = heap := {num_entries = num_entries + 1,  
		     cleanup_thres = cleanup_thres,
		     pq = pq }

    val _ =  percolateUp (x, pq , num_entries )
  in
    ()
  end

fun findMin heap =
  let
    val {num_entries, pq, cleanup_thres } = !heap

    val _ = if (Priority.compare(Array.sub(pq, 0), Priority.EMPTY) = EQUAL) then
	      (print "findMin failed: priority queue is empty\n"; raise InternalError_BinaryHeap)
	    else ()

  in
    if (num_entries > 0) then
      SOME (Array.sub(pq, 0))
    else
      NONE
  end

  fun collectGarbage heap = 
    let
      val {num_entries, cleanup_thres, pq} = !heap 
    in
      if num_entries >= cleanup_thres then
	renewHeap (heap, false)
      else ()
    end

  fun deleteMin heap =
    let
      val minVal = findMin heap
      (* Swap last entry with the first *)
      val {num_entries, pq, cleanup_thres} = !heap 
      val lastEntry = Array.sub(pq, num_entries - 1)
      val _ = Array.update (pq, 0, lastEntry)
      val thr = cleanup_thres

      val _ = heap := {num_entries = num_entries - 1, 
		       cleanup_thres = thr, pq = pq}

      val _ = percolateDown (num_entries - 1, pq, 0)
    in
      minVal
    end
    
  fun discardMin heap = (deleteMin heap; ())

end
