(**********************************************************************
 ** TestListInputApp.sml
 **
 ** Copyright 2007 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: 
 **********************************************************************)

functor TestListInputApp 
   (structure InElt : ELT
    structure InNAElt : ELT
    val convertElt : InElt.t -> InNAElt.t
    structure Generator : GENERATOR 
       where type t = InElt.t list
    structure OutType :
       sig
          structure Adaptive : TYPE
          structure NonAdaptive : TYPE
          structure Checker : CHECKER
             where type a_t = Adaptive.t
             where type na_t = NonAdaptive.t
       end
    structure AdaptiveAutoApp : APPLICATION
       where type it = InElt.t ListAdaptive.t
       where type ot = OutType.Adaptive.t
    structure NonAdaptiveAutoApp : APPLICATION
       where type it = InNAElt.t ListNonAdaptive.t
       where type ot = OutType.NonAdaptive.t
    structure NonAdaptiveUnAutoApp : APPLICATION
       where type it = InNAElt.t ListNonAdaptive.t
       where type ot = OutType.NonAdaptive.t) =
struct

structure ListInType = BuildListInOut
			 (structure Elt = InElt
                          structure NAElt = InNAElt
		          val convertElt = convertElt
			 )

structure TestApp =
   TestApp
   (structure InType = ListInType
    structure Generator = 
       struct
          type t = InElt.t ListAdaptive.t
          fun gen {size, seed} =
             ListMetaAdaptive.fromListEq
             (InElt.eq, Generator.gen {size = size, seed = seed})
       end 
    structure OutType = OutType
    structure AdaptiveAutoApp = AdaptiveAutoApp
    structure NonAdaptiveAutoApp = NonAdaptiveAutoApp
    structure NonAdaptiveUnAutoApp = NonAdaptiveUnAutoApp)
open TestApp

(**********************************************************************
 ** Mainline
 **********************************************************************)
(*
structure Simulator = SimulateMotion
*)

val SEED = ref 56720
fun newList n = ((SEED := 1 + !SEED); Generator.gen {size = n, seed = !SEED})

fun testInitialRun' (n,rounds,flag) = 
  let
    val _ = AdaptiveMeta.init ()

    val str = "\ntestInitialRun(" ^ Int.toString(n) ^ ")"
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    val l0 = newList n
    val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l0)
    val is0 = iToString(i0)


    fun experiment () = 
    let
      val _ = doIf verbose (fn () => print "\nInitial Run:")
      val r0 = AdaptiveAutoApp.run i0
      val rs0 = oToString(!r0)

      (* Check the answer *)
      val test = checkResult (i0,!r0) "Initial Run"

      val _ = passFail (test)

      val _ = doIf verbose (fn () => pPrint ("\nInput = " ^ is0 ^
 	   		                     "\nResult = " ^ rs0 ^ "\n"))
     in
        ("", test)
     end
     
  in
    runExperiment (flag, "testInitialRun") (experiment,rounds)
  end

(**********************************************************************
 ** Test a single insertion and deletion.  
 ** Builds a list of n elements, takes out the first key, sorts the
 ** rest, and inserts/deletes the key taken out, at the given
 ** position.  Takes linear time to find the position.  
 **********************************************************************)

fun testInsertDelete' (n,rounds,insertPosition,flag) =
  let

    val _ = AdaptiveMeta.init ()

    val str = ("\ntestInsertDelete(" ^ Int.toString(n) ^ "," ^ Int.toString(insertPosition) ^ ")")
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    val l = newList n
    val h = hd l 
    val t = tl l
    val i0 = ListMetaAdaptive.fromListEq (InElt.eq, t)


    fun experiment () = 
      let
        (** Initial run **)
        val _ = doIf verbose (fn () => print ("\nInitial Run"))
        val is0 = iToString i0
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)

       
        (* Insert an element  *)
        val _ = doIf verbose (fn () => print "\nRerun After Insert")
        val _ = ListMetaAdaptive.insertAtEq (InElt.eq, ListMetaAdaptive.drop (i0, insertPosition), h)
        val is1 = iToString(i0)
        val _ = AdaptiveMeta.propagate()
        val _ = doIf verbose (fn () => print "\nPropagate  done.")
        val rs1 = oToString(!r0)

        (* Check the answer *)
        val test1 = checkResult (i0,!r0) "Insert"

        (* Delete the element  *)
        val _ = doIf verbose (fn () => print "\nRerun After Delete")
        val _ = ListMetaAdaptive.deleteFirstAt (i0, fn e => InElt.eq (e,h))
        val is2 = iToString(i0)
        val _ = AdaptiveMeta.propagate()
        val _ = doIf verbose (fn () => print "\nPropagate  done.")
        val rs2 = oToString(!r0)

        (* Check the answer *)
        val test2 = checkResult (i0,!r0) "Delete"
    

        val test =  test1 andalso test2
        val _ = passFail (test)

        val _ = doIf verbose (fn () => pPrint ("\nInput = " ^ is0 ^
    			                       "\nResult = " ^ rs0 ^
    			                       "\nInput = " ^ is1 ^
			                       "\nResult = " ^ rs1 ^
			                       "\nInput = " ^ is2 ^
			                       "\nResult = " ^ rs2))
        in
	  ("", test)
        end
  in
    runExperiment (flag, "testInsertDelete") (experiment,rounds)
  end

(**********************************************************************
 ** testAllInsertDelete
 **********************************************************************)
fun testAllInsertDelete' (n,rounds,flag) =
  let

    val _ = AdaptiveMeta.init ()

    val str = ("\ntestAllInsertDelete(" ^ Int.toString(n) ^  ")")
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    val l = newList n
    val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l)
    val ls0 = iToString i0

    fun experiment () = 
      let

        (** Initial run **)
        val _ = doIf verbose (fn () => print ("\nInitial Run"))
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)
        val _ = doIf verbose (fn () => print "initial run done\n")


        fun deleteInsertAll l =
          let
    	    fun deleteFirst l =
	      let val hcell = AdaptiveMeta.deref l
	      in 
                case hcell of
	          ListAdaptive.NIL => raise BadCode 
	        | ListAdaptive.CONS(h,t) => (ListMetaAdaptive.change (l, AdaptiveMeta.deref t); (hcell,t))
	      end
          in
    	    case (AdaptiveMeta.deref l) of
	      ListAdaptive.NIL => true
	    | _ =>
	      let
	        (* Delete an element  *)
                val _ = doIf verbose (fn () => print "Deleting\n")
	        val (hcons,t) = deleteFirst l
	        val _ = AdaptiveMeta.propagate ()
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls1 = iToString i0
	        val rs1 = oToString (!r0)
                val test1 = checkResult (i0,!r0) "Delete"
		
	        (* Insert the element back  *)
                val _ = doIf verbose (fn () => print "Inserting\n")
	        val _ = ListMetaAdaptive.change (l, hcons)
	        val _ = AdaptiveMeta.propagate ()
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls2 = iToString i0
	        val rs2 = oToString (!r0)
                val test2 = checkResult (i0,!r0) "Insert"

		val _ = doIf verbose (fn () => pPrint ("\nInput = " ^ ls0 ^
    						       "\nResult = " ^ rs0 ^
    						       "\nInput = " ^ ls1 ^
						       "\nResult = " ^ rs1 ^
						       "\nInput = " ^ ls2 ^
						       "\nResult = " ^ rs2))



  	      in
	        test1 andalso test2 andalso (deleteInsertAll t)
	      end
        end
      val test = deleteInsertAll i0
      val _ = passFail (test)
    in
      ("", test)
    end
  in
    runExperiment (flag, "testAllInsertDelete") (experiment,rounds)
  end

(**********************************************************************
 ** testAllInsertDeleteK
 ** Like testAllInsertDelete but performs k deletions at a time. 
 **********************************************************************)
fun testAllInsertDeleteK' (n,k,rounds,flag) =
  let
    val _ = AdaptiveMeta.init ()

    val str = ("\ntestAllInsertDelete(" ^ Int.toString(n) ^  ")")
    val (verbose,ItoString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    val l = newList n
    val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l)
    val ls0 = ItoString i0

    fun experiment () = 
      let

        (** Initial run **)
        val _ = doIf verbose (fn () => print ("\nInitial Run"))
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)
        val _ = doIf verbose (fn () => print "initial run done\n")


        fun deleteInsertAll l =
          let
	    fun  take (l,k) = 
              if k = 0 then  l 
	      else case (AdaptiveMeta.deref l) of
		       ListAdaptive.NIL => l
		     | ListAdaptive.CONS(_,t) => take (t,k-1)
          in
    	    case (AdaptiveMeta.deref l) of
	      ListAdaptive.NIL => true
	    | cFirst =>
	      let
	        (* Delete the first k elements  *)
                val _ = doIf verbose (fn () => print "Deleting\n")
       
	        val t_k = take (l,k)
                val _ = ListMetaAdaptive.change (l, AdaptiveMeta.deref t_k)
	        val _ = AdaptiveMeta.propagate ()
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls1 = ItoString i0
	        val rs1 = oToString (!r0)
                val test1 = checkResult (i0,!r0) "Delete"
		
	        (* Insert the elements back  *)
                val _ = doIf verbose (fn () => print "Inserting\n")
	        val _ = ListMetaAdaptive.change (l, cFirst)
	        val _ = AdaptiveMeta.propagate ()
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls2 = ItoString i0
	        val rs2 = oToString (!r0)
                val test2 = checkResult (i0,!r0) "Insert"

		val _ = doIf verbose (fn () => pPrint ("\nInput = " ^ ls0 ^
    						       "\nResult = " ^ rs0 ^
    						       "\nInput = " ^ ls1 ^
						       "\nResult = " ^ rs1 ^
						       "\nInput = " ^ ls2 ^
						       "\nResult = " ^ rs2))

  	      in
	        test1 andalso test2 andalso (deleteInsertAll t_k)
	      end
        end
      val test = deleteInsertAll i0
      val _ = passFail (test)
    in
      ("", test)
    end
  in
    runExperiment (flag, "testAllInsertDeleteK") (experiment,rounds)
  end

(** Delete and insert k elements (equally spaced) starting at each element **)
fun testBatch' (n,k,rounds,flag) =
  let
    val _ = AdaptiveMeta.init ()

    val k = if n < k then (print "WARNING: the list is to small setting the batch size to the input size\n";  n) else k

    val str = ("\ntestBatch(" ^ Int.toString(n) ^ ", " ^ Int.toString(k) ^  ")")
    val (verbose,ItoString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    fun toArray l = 
      let
        fun toList l = 
          case (AdaptiveMeta.deref l) of
            ListAdaptive.NIL => nil
          | ListAdaptive.CONS(_,t) => l::toList t
    
      in Array.fromList (toList l) end

    val m = Int.div (n,k)  
    val l = newList n
    val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l)
    val ls0 = ItoString i0
    val a = toArray i0  


(*    val _ = print ("m = " ^ Int.toString m ^ "\n") *)
    fun experiment () = 
      let

        (** Initial run **)
        val _ = doIf verbose (fn () => print ("\nInitial Run"))
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)
        val _ = doIf verbose (fn () => print "initial run done\n")


        fun deleteInsertAll i =
          let
	    fun  take (cells,i,j) = 
              if (j = 0 orelse i >= n) then  cells
	      else 
                let
		  val nexti = Int.mod(i + m, n)
		  val m = Array.sub (a,i)
		  val (c as ListAdaptive.CONS(_,t)) = AdaptiveMeta.deref m
                  val _ = ListMetaAdaptive.change (m, AdaptiveMeta.deref t)
(*                  val _ = print (Int.toString i  ^ ", ") *)
		in
		  take ((i,c)::cells, nexti, j-1)
		end

            fun insert cells = 
		case cells of
		  nil => ()
		| (i,c)::tcells => 
		  let 
		    val m = Array.sub (a,i)
		    val _ = ListMetaAdaptive.change (m, c)
		  in
		    insert tcells
		  end
          in
            if (i = m) then 
	      true
	    else
	      let
	        (* Delete equally spaced k elements  *)
                val _ = doIf verbose (fn () => print "Deleting\n")
(*                val _ = print ("I = " ^ Int.toString i ^ "\n") *)
	        val cells = take (nil,i,k)
(*                val _ = print "\n"		     *)
(*		val _ = inc (count) *)
        	val _ = AdaptiveMeta.propagate () 
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls1 = ItoString i0
	        val rs1 = oToString (!r0)
                val test1 = checkResult (i0,!r0) "Delete"
		
	        (* Insert the elements back  *)
                val _ = doIf verbose (fn () => print "Inserting\n")
	        val _ = insert cells
(*		val _ = inc (count) *)
	        val _ = AdaptiveMeta.propagate () 
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls2 = ItoString i0
	        val rs2 = oToString (!r0)
                val test2 = checkResult (i0,!r0) "Insert"

		val _ = doIf verbose (fn () => pPrint ("\nInput = " ^ ls0 ^
    						       "\nResult = " ^ rs0 ^
    						       "\nInput = " ^ ls1 ^
						       "\nResult = " ^ rs1 ^
						       "\nInput = " ^ ls2 ^
						       "\nResult = " ^ rs2))
  	      in
	        test1 andalso test2 andalso (deleteInsertAll (i+1))
	      end
        end
      val test = deleteInsertAll 0
      val _ = passFail (test)
    in
      ("", test)
    end
  in
    runExperiment (flag, "testAllInsertDeleteK") (experiment,rounds)
  end



fun testLocal' (n,k,rounds,flag) =
  let
    val _ = AdaptiveMeta.init ()

    val k = if n < k then (print "WARNING: the list is to small setting the subsize to the input size\n";  n) else k
    val str = ("\ntestLocal(" ^ Int.toString(n) ^ ", " ^ Int.toString(k) ^  ")") 
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    fun ithTail (i,l) = 
     if i = 0 then
       l
     else 
       case (AdaptiveMeta.deref l) of
	 ListAdaptive.NIL => raise BadCode
       | ListAdaptive.CONS(h,t) => ithTail (i-1,t)

    val l = newList n
    val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l)
    val ls0 = iToString i0

    fun experiment () = 
      let
        (** Initial run **)
        val _ = doIf verbose (fn () => print ("\nInitial Run"))
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)
        val _ = doIf verbose (fn () => print "initial run done\n")


        fun deleteInsertAll l k =
          let
    	    fun deleteFirst l =
	      let val hcell = AdaptiveMeta.deref l
	      in 
                case hcell of
	          ListAdaptive.NIL => raise BadCode 
	        | ListAdaptive.CONS(h,t) => (ListMetaAdaptive.change (l, AdaptiveMeta.deref t); (hcell,t))
	      end
          in
            if k = 0 then
	      true
	    else
    	      case (AdaptiveMeta.deref l) of
	      ListAdaptive.NIL => true
	    | _ =>
	      let
(*		val _ = print ("k = " ^ Int.toString k ^ "\n") *)
	        (* Delete an element  *)
                val _ = doIf verbose (fn () => print "Deleting\n")
	        val (hcons,t) = deleteFirst l
	        val _ = AdaptiveMeta.propagate ()
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls1 = iToString i0
	        val rs1 = oToString (!r0)
                val test1 = checkResult (i0,(!r0)) "Delete"
		
	        (* Insert the element back  *)
                val _ = doIf verbose (fn () => print "Inserting\n")
	        val _ = ListMetaAdaptive.change (l, hcons)
	        val _ = AdaptiveMeta.propagate ()
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls2 = iToString i0
	        val rs2 = oToString (!r0)
                val test2 = checkResult (i0,(!r0)) "Insert"

		val _ = doIf verbose (fn () => pPrint ("\nInput = " ^ ls0 ^
    						       "\nResult = " ^ rs0 ^
    						       "\nInput = " ^ ls1 ^
						       "\nResult = " ^ rs1 ^
						       "\nInput = " ^ ls2 ^
						       "\nResult = " ^ rs2))



  	      in
	        test1 andalso test2 andalso (deleteInsertAll t (k-1))
	      end
        end
      val m = Int.div (n-k,2) 
      val t = Int.div (n,k)
      val l = ithTail (m,i0)
(*
      val _ = print ("subsize = " ^ Int.toString k ^ 
		     "repeat = " ^ Int.toString t ^ 
		     "start = " ^ Int.toString m ^ "\n")
*)
      val test = repeatTest (fn () => deleteInsertAll l k) t
      val _ = passFail test

    in
      ("", test)
    end
  in
    runExperiment (flag, "testLocal") (experiment,rounds)
  end



(**********************************************************************
 ** testChangeAll
 **********************************************************************)
fun testChangeAll' f (n,rounds,flag) =
  let

    val _ = AdaptiveMeta.init ()

    val str = ("\ntestChangeAll(" ^ Int.toString(n) ^  ")")
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    val l = newList n
    val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l)
    val ls0 = iToString i0

    fun experiment () = 
      let
        (** Initial run **)
        val _ = doIf verbose (fn () => print ("\nInitial Run"))
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)
        val _ = doIf verbose (fn () => print "initial run done\n")

        fun changeAll l =
	  let
          in
    	    case (AdaptiveMeta.deref l) of
	      ListAdaptive.NIL => true
	    | hcell as ListAdaptive.CONS(h,t) =>
	      let
	        (* Change the head  *)
                val _ = doIf verbose (fn () => print "Changing\n")
	        val _ = ListMetaAdaptive.change (l, ListAdaptive.CONS (f h, t))
	        val _ = AdaptiveMeta.propagate ()
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls1 = iToString i0
	        val rs1 = oToString (!r0)
                val test1 = checkResult (i0,(!r0)) "Change"

	        (* Change it back  *)
(*
                val _ = doIf verbose (fn () => print "Changing back\n")
	        val _ = ML.change l hcell 
	        val _ = AdaptiveMeta.propagate ()
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls2 = iToString i0
	        val rs2 = oToString r0
                val test2 = checkResult (i0,r0) "Change"

		val _ = doIf verbose (fn () => pPrint ("\nInput = " ^ ls0 ^
    						       "\nResult = " ^ rs0 ^
    						       "\nInput = " ^ ls1 ^
						       "\nResult = " ^ rs1 ^
						       "\nInput = " ^ ls2 ^
						       "\nResult = " ^ rs2))
*)
  	      in
	        test1 andalso (changeAll t)
	      end
        end
      val test = changeAll i0
      val _ = passFail (test)
    in
      ("", test)
    end
  in
    runExperiment (flag, "testChangeAll") (experiment,rounds)
  end



fun testChangeAllK' f (n, k, rounds, flag) =
  let
    val _ = AdaptiveMeta.init ()

    val k = if n < k then (print "WARNING: the list is to small setting the batch size to the input size\n";  n) else k

    val str = ("\ntestChangeAllK(" ^ Int.toString(n) ^ ", " ^ Int.toString(k) ^ ")")
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    fun toArray l = 
      let
        fun toList l = 
          case (AdaptiveMeta.deref l) of
            ListAdaptive.NIL => nil
          | ListAdaptive.CONS(_,t) => l::toList t
    
      in Array.fromList (toList l) end

    val m = Int.div (n,k)  
    
    val l = newList n
    val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l)
    val ls0 = iToString i0
    val a = toArray i0  

(*    val _ = print ("m = " ^ Int.toString m ^ "\n") *)
    fun experiment () = 
      let
        (** Initial run **)
        val _ = doIf verbose (fn () => print ("\nInitial Run"))
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)
        val _ = doIf verbose (fn () => print "initial run done\n")


        fun changeAll i =
          let
	    fun  take (i,j) = 
              if (i >= n orelse j = 0) then  ()
	      else 
                let
                  val _ = doIf verbose (fn () => print ("A:" ^ Int.toString i  ^ ", "))
		  val nexti = Int.mod(i + m, n)
		  val m = Array.sub (a,i)
		  val (ListAdaptive.CONS(h,t)) = AdaptiveMeta.deref m
	        val _ = ListMetaAdaptive.change (m, ListAdaptive.CONS (f h, ListMetaAdaptive.newEq (InElt.eq, AdaptiveMeta.deref t)))
                  val _ = doIf verbose (fn () => print (Int.toString i  ^ ", "))
		in
		  take (nexti, j-1)
		end
          in
            if (i = m) then 
	      true
	    else
	      let
	        (* Change equally spaced k elements  *)
                val _ = doIf verbose (fn () => print "Changing\n")
                val _ = doIf verbose (fn () => print ("I = " ^ Int.toString i ^ "\n") )
	        val cells = take (i,k)
(*                val _ = print "\n"		     *)
(*		val _ = inc (count) *)
                val _ = doIf verbose (fn () => print "Starting Propagate\n")

        	val _ = AdaptiveMeta.propagate () 
                val _ = printCount 100 
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls1 = iToString i0
	        val rs1 = oToString (!r0)
                val test1 = checkResult (i0,(!r0)) "Change"

		val _ = doIf verbose (fn () => pPrint ("\nInput = " ^ ls0 ^
    						       "\nResult = " ^ rs0 ^
    						       "\nInput = " ^ ls1 ^
						       "\nResult = " ^ rs1))
  	      in
	        test1 andalso (changeAll (i+1))
	      end
        end
(*      val _ = reset count *)
      val test = changeAll 0
(*      val _ = print ("count = " ^ Int.toString (!count) ^ "\n") *)
      val _ = passFail (test)
    in
      ("", test)
    end
  in
    runExperiment (flag, "testChangeAllK") (experiment,rounds)
  end


fun testChangeAllK' f (n, k, rounds, flag) =
  let
    val _ = AdaptiveMeta.init ()

    val k = if n < k then (print "WARNING: the list is to small setting the batch size to the input size\n";  n) else k

    val str = ("\ntestChangeAllK(" ^ Int.toString(n) ^ ", " ^ Int.toString(k) ^ ")")
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    val l = newList n
    val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l)
    val ls0 = iToString i0

    fun experiment () = 
      let
        (** Initial run **)
        val _ = doIf verbose (fn () => print ("\nInitial Run"))
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)
        val _ = doIf verbose (fn () => print "initial run done\n")


        fun changeAll l =
          let
 
	    fun  loop (l,j) = 
              let
                val _ = print ("j = " ^ Int.toString j ^ "\n") 
	      in
		 if (j = 0) then  (print ("j = " ^ Int.toString j ^ "\n") ; l)
		 else
		     case (AdaptiveMeta.deref l) of
			 ListAdaptive.NIL => l
		       | ListAdaptive.CONS (h,t) => 
(*			 let val _ = ML.change l (ListAdaptive.CONS (f h, t))  
                         this works too but it is  imperative and probably slower *) 

			 let val _ = ListMetaAdaptive.change (l, ListAdaptive.CONS (f h, ListMetaAdaptive.newEq (InElt.eq, AdaptiveMeta.deref t))) 
			 in loop (t, j-1)
			 end
	      end
          in
	      case (AdaptiveMeta.deref l) of
	        ListAdaptive.NIL => true
	      | _ => 
	      let
	        (* Change equally spaced k elements  *)
                val _ = doIf verbose (fn () => print "Changing\n")
                val _ = doIf verbose (fn () => print ("k = " ^ Int.toString k ^ "\n"))
	        val l' = loop (l,k)
	        val ls1 = iToString i0

		val _ = doIf verbose (fn () => pPrint ("\nInput = " ^ ls0 ^
    						       "\nInput = " ^ ls1 ^ "\n"))

                val _ = doIf verbose (fn () =>  print "Propagating... \n")
(*		val _ = inc (count) *)
        	val _ = AdaptiveMeta.propagate () 
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls1 = iToString i0
	        val rs1 = oToString (!r0)
                val test1 = checkResult (i0,(!r0)) "Change"

		val _ = doIf verbose (fn () => pPrint ("\nInput = " ^ ls0 ^
    						       "\nResult = " ^ rs0 ^
    						       "\nInput = " ^ ls1 ^
						       "\nResult = " ^ rs1))
  	      in
	        test1 andalso (changeAll l')
	      end
        end
(*      val _ = reset count *)
      val test = changeAll i0
(*      val _ = print ("count = " ^ Int.toString (!count) ^ "\n") *)
      val _ = passFail (test)
    in
      ("", test)
    end
  in
    runExperiment (flag, "testChangeAllK") (experiment,rounds)
  end



  (**********************************************************************
   ** Incremental
   ** Incrementally sort the list by adding keys at the end
   **********************************************************************)

(* Incrementally sort l by adding each key of l at the _end_ of ml (presumed to be empty) *)
fun multiInsert (ml,l,flag) =
  let
    fun MI(mlEnd,l,i) =
      case l of
	nil => ()
      | h::t => 
	  let
(*            val _ = print ("multiInsert: i = " ^ Int.toString i ^ "\n") *)
	    val mlEnd' = ListMetaAdaptive.insertAtEq (InElt.eq,mlEnd,h)
	    val _ = AdaptiveMeta.propagate()
	  in
	    MI(mlEnd',t,i+1)
	  end

  in
    MI(ml,l,0)
  end


fun multiInsertK k (ml,l,flag) =
  let
    fun take (mlEnd, l, k) = 
	if (k = 0) then
	  (mlEnd, l)
	else
	  case l of
	    nil => (mlEnd, l)
	  | h::t => 
	    let
(*	      val _ = inc (count) *)
	      val mlEnd' = ListMetaAdaptive.insertAtEq (InElt.eq,mlEnd,h)
	    in
	      take (mlEnd', t, k-1)
	    end

    fun MI(mlEnd,l) =
	case l of
	  nil => ()
	| _=> 
	  let
	    val (mlEnd',l') = take (mlEnd, l, k)
	    val _ = AdaptiveMeta.propagate ()
	  in
	    MI (mlEnd',l')
	  end
  in	  
    MI(ml,l)
  end


(* Delete whole list from the beginning one at a time *)
fun multiDelete (ml,flag) =
  let
    fun MD(ml,i) =
      case AdaptiveMeta.deref ml of
	ListAdaptive.NIL => ()
      | ListAdaptive.CONS(h,t) => 
	  (MD(t,i+1);
   	   doIf flag (fn () => print ("\nDelete " ^ Int.toString i));
	   ListMetaAdaptive.deleteAt ml;
	   AdaptiveMeta.propagate(); ())
  in
    MD(ml,0)
  end

fun testMultiDeleteInsert' (n,rounds,flag) =
  let
    val _ = AdaptiveMeta.init ()

    val str = ("\ntestMultiDeleteInsert(" ^ Int.toString(n) ^ ")")
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)


    val l0 = newList n
    val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l0)
    val is0 = iToString(i0)


    fun experiment () = 
      let
        val _ = doIf verbose (fn () => print "\nInitial Run:")
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)

        (** Multi delete **)
        val _ = doIf verbose (fn () => print "\n Multi delete:")
        val _ = multiDelete (i0,verbose)
        val rs1 = oToString (!r0)
   
        val test1 = checkResult (i0,(!r0)) "Delete All"

        (** Multi insert **)
        val _ = doIf verbose (fn () => print "\n Multi Insert:")
        val _ = multiInsert (i0,l0,verbose)
        val rs2 = oToString (!r0)

        val test2 = checkResult (i0,(!r0)) "Insert All"


        val test = test1 andalso test2 
       
        val _ = passFail test

        val _ = doIf verbose (fn () => pPrint ("\nInput                         = " ^ is0 ^  
                                              "\nResult (initial)              = " ^ rs0 ^       
                                              "\nResult (after delete all)     = " ^ rs1 ^       
                                              "\nResult (after insert all)     = " ^ rs2 ))
      in
        ("", test)
      end

  in
    runExperiment (flag, "testMultiDeleteInsert") (experiment,rounds)
  end

fun testMultiInsert' (n,rounds, flag) =
  let

    val _ = AdaptiveMeta.init ()

    val str = ("\ntestMultiDeleteInsert(" ^ Int.toString(n) ^ ")")
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    val l0 = newList n

    fun experiment () = 
      let
	val i0 = ListMetaAdaptive.fromListEq (InElt.eq, [])
	val is0 = iToString(i0)

        val _ = doIf verbose (fn () => print "\nInitial Run:")
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)

        val _ = doIf verbose (fn () => print "\n Multi Insert:")
        val _ = multiInsert (i0,l0,verbose)
        val test = checkResult (i0,(!r0)) "Multi Insert"
        val _ = passFail test
        val rs1 = oToString (!r0)

        val _ = doIf verbose (fn () => pPrint ("\nInput                 = " ^ is0 ^  
                                              "\nResult (initial)      = " ^ rs0 ^       
                                              "\nResult                = " ^ rs1 ))
      in
        ("", test)
      end

  in
    runExperiment (flag, "testMultiInsert") (experiment,rounds)
  end


fun testMultiInsertBatch' (n, k, rounds, flag) =
  let

    val _ = AdaptiveMeta.init ()

    val str = ("\ntestMultiDeleteInsert(" ^ Int.toString(n) ^ ")")
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)


    val l0 = newList n

    fun experiment () = 
      let
	val i0 = ListMetaAdaptive.fromListEq (InElt.eq, [])
	val is0 = iToString(i0)

        val _ = doIf verbose (fn () => print "\nInitial Run:")
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)

        val _ = doIf verbose (fn () => print "\n Multi Insert:")
(*        val _ = reset (count) *)
        val _ = multiInsertK k (i0,l0,verbose)
(*        val _ = TextIO.print ("count = " ^ Int.toString (!count) ^ "\n")  *)
        val test = checkResult (i0,(!r0)) "Multi Insert"
        val _ = passFail test
        val rs1 = oToString (!r0)

        val _ = doIf verbose (fn () => pPrint ("\nInput                 = " ^ is0 ^  
                                              "\nResult (initial)      = " ^ rs0 ^       
                                              "\nResult                = " ^ rs1 ))
      in
        ("", test)
      end

  in
    runExperiment (flag, "testMultiInsert") (experiment,rounds)
  end


(*
fun testKinetic' eltAtTime (n, nEvents, flag) = 
  let

    val _ = AdaptiveMeta.init ()

    val _  = Simulator.init()

    val str = ("\ntestKinetic!(" ^ Int.toString(n) ^ ")")
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    fun  ItoStaticKeyList time i = List.map (fn p => eltAtTime (p, time)) (ListMetaAdaptive.toList i)

    val l0 = newList n
    val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l0)
    val is0 = iToString(i0)

    val _ = doIf verbose (fn () => print "\nInitial Run:")
    val r0 = AdaptiveAutoApp.run i0
    val rs0 = oToString (!r0)
    val _ = doIf verbose (fn () => pPrint ("\nInput                 = " ^ is0 ^  
                                           "\nResult (initial)      = " ^ rs0))
    val _ =  checkResult (i0,!r0) "Kinetic/InitialRun"
(** ? 
    val _ = ModResult.checkForExtEvent (Scheduler.EXTERNAL_EVENT_FLAG) r0
**)
    fun experiment () = 
      let

        val _ = doIf verbose (fn () => print "\n Starting the kinetic event scheduler:")

        (* No need for time here actually, the applications access it directly **)
        fun testAtTime time = 
          let
	    val _ = doIf verbose (fn () =>  pPrint ("app output = "^(oToString (!r0))^"\n"))

            val test =  checkResult (i0,!r0) "Kinetic/InitialRun"
	  in
	    (passFail test; test)
          end

        val ((nEvents, nExtEvents, nFlips, maxGap, avGap), test) = 
	    doIfElse (areWeTesting flag)
		     (fn () => Simulator.testSimulate testAtTime )
		     (fn () => Simulator.simulate nEvents)
     
      in
        ("\n" ^ "nEvents="^(Int.toString nEvents) ^ "\n" ^ 
	        "nExtEvents="^(Int.toString nExtEvents) ^ "\n" ^ 
	        "nFlips="^(Int.toString nFlips) ^ "\n" ^ 
                "maximumGap=" ^ Real.toString maxGap ^ "\n" ^ 
	        "averageGap=" ^ Real.toString avGap ^ "\n",
	 test)
      end

    val resp = runExperiment (flag, "testKinetic") (experiment,1)
    val _ = Simulator.init()
  in
    resp
  end

(**
fun testKineticRandom' (n, nEvents, flag) = 
  let
    infix ==> val op ==> = Comb.read

    val _ = Meta.init ()

    val _  = Scheduler.init()

    val str = ("\ntestKinetic(" ^ Int.toString(n) ^ ")")
    val (verbose,iToString,oToString,pPrint,checkResult,passFail,getTimeStamps) = initExperiment (flag,str)
    val testing = flag = TEST_VERBOSE orelse flag = TEST_SUMMARY

    fun  ItoStaticKeyList time i = 
         List.map (fn p => IElt.atTime (p, time)) (ModList.toList i)

    val l0 = newList n
    val i0 = ML.fromList l0
    val is0 = iToString(i0)

    val _ = doIf verbose (fn () => print "\nInitial Run:")
    val r0 = App.run i0
    val rs0 = oToString r0
    val _ = doIf verbose (fn () => pPrint ("\nInput                 = " ^ is0 ^  
                                           "\nResult (initial)      = " ^ rs0))


    val _ = ModResult.checkForExtEvent (Scheduler.EXTERNAL_EVENT_FLAG) r0

    fun experiment () = 
      let

        val _ = doIf verbose (fn () => print "\n Starting the kinetic event scheduler:")
        val _ = doIf verbose (fn () =>
		     let 
			 val time = Scheduler.getTime()
		         val vresult = Verifier.run (ItoStaticKeyList time i0)
                         val test = Checker.eq (r0, vresult, time)
                     in
                         print ("\nverifier result = " ^  (VResult.toString vresult))
                     end)

        (* Something like this function needs to be fed into testSimulate *)
	val state = Random.rand(437289,5478)
        fun testAtTime (l,r) time = 
          if Random.randReal state < 0.80 then
	      true
	  else
              let
		  val _ = doIf verbose (fn () =>  pPrint ("app output = "^(oToString r)^"\n"))

		  val vresult = Verifier.run (ItoStaticKeyList time l)
		  val test = Checker.eq (r, vresult, time)
              in
		  (if (verbose andalso not test) then
		       pPrint("\n App Result      = " ^(oToString r)^
                              "\n Verifier Result = " ^(VResult.toString vresult))
		   else
		       ()
		  );
		  test
              end

        val ((nEvents, nExtEvents, nFlips, maxGap, avGap), test) = doIfElse testing (fn () => Scheduler.testSimulate (testAtTime (i0,r0)) )
							       (fn () => Scheduler.simulate nEvents)
     
      in
        ("\n" ^ "nEvents="^(Int.toString nEvents) ^ "\n" ^ 
	        "nExtEvents="^(Int.toString nExtEvents) ^ "\n" ^ 
	        "nFlips="^(Int.toString nFlips) ^ "\n" ^ 
                "maximumGap=" ^ Real.toString maxGap ^ "\n" ^ 
	        "averageGap=" ^ Real.toString avGap ^ "\n",
	 test)
      end

  in
    case flag of
	TEST_SUMMARY => 
	let 
	    val resp = TimeProfile.timeN 1  "testKinetic" (fn () => experiment ())
	    val _ = Scheduler.init()
	in
	    resp
	end
      | TEST_VERBOSE =>
	let 
	    val resp = TimeProfile.timeN 1  "testKinetic" (fn () => experiment ())
	    val _ = Scheduler.init()
	in
	    resp
	end
      | PROFILE_TIME => 
	let
	    val resp = TimeProfile.profileN 1 (fn () => experiment ())
	    val _ = Scheduler.init()
	in
	    resp
	end
      | PROFILE_SPACE => 
	let
	    val resp = TimeProfile.profileN 1 (fn () => experiment ())
	    val _ = Scheduler.init()
	in
	    resp
	end			       
  end


**)
out kinetic stuff
**)

(**********************************************************************
 ** Testing Functions
 **********************************************************************)

fun testInitialRun(n) = testInitialRun' (n,1,(TEST_FLAG))
fun testInsertDelete (n,i) = testInsertDelete' (n,1,i,(TEST_FLAG))

fun testFront(n) = testInsertDelete(n, 0)
fun testMid(n) = testInsertDelete(n, n div 2)
fun testEnd(n) = testInsertDelete(n, n-1)

fun testMultiInsert (n) = 
  testMultiInsert'(n,1,(TEST_FLAG))

fun testMultiInsertBatch (n,k) = 
  testMultiInsertBatch'(n,k,1,(TEST_FLAG))

fun testMultiDeleteInsert (n) =  
  testMultiDeleteInsert'(n,1,(TEST_FLAG))

fun testAllInsertDelete (length) =
  testAllInsertDelete' (length,1,(TEST_FLAG))

fun testAllInsertDeleteK (length,k) =
  testAllInsertDeleteK' (length,k,1,(TEST_FLAG))

fun testBatch (length,k) = 
  testBatch' (length,k,1,TEST_FLAG)

fun testLocal (length,k) = 
  testLocal' (length,k,1,TEST_FLAG)


fun testChangeAll  f length =
  testChangeAll' f (length,1,(TEST_FLAG))

fun testChangeAllK  f (length,k) =
  testChangeAllK' f (length,k, 1,(TEST_FLAG))

(*
fun testKinetic f (length,k) = testKinetic' f (length, 0, TEST_FLAG)
*)
fun testAll' verbose n = 
  (testInitialRun' (n,1,verbose) andalso

   testInsertDelete'(n,1, n-1, verbose) andalso
   testInsertDelete'(n, 1, n div 2, verbose) andalso 
   testInsertDelete'(n, 1, 0, verbose) andalso 

(*   testMultiDeleteInsert'(n, 1, verbose)  andalso  *)
   testAllInsertDelete' (n, 1, verbose) andalso
   testBatch' (n, 2, 1, verbose)
  )

val testAll = testAll' (TEST_FLAG)

fun testRange test (min,max) = 
  let
    fun f(n) =
      if (n=max) then true
      else (print ("Test: " ^ Int.toString n ^ "\n");
	    (test n) andalso f(n+1))
  in
    f(min)
  end

fun testAllUpTo (n) = testRange testAll (1,n)
fun testAllUpToFrom (m,n) = testRange testAll (m,n)


(**********************************************************************
 ** Profiler Functions
 **********************************************************************)

fun profileVerifierCPS (n,rounds) = 
  let
    val _ = print ("ProfileVerifierCPS: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
    fun f () = 
      let val l = newList n
	  val l = ListMetaAdaptive.fromListEq (InElt.eq, l)
          val l = ListInType.Converter.convert l
      in NonAdaptiveAutoApp.run l;("",true) end
  in
    profileN "profileVerifierCPS" (f,rounds)
  end
fun profileVerifier (n,rounds) = 
  let
    val _ = print ("ProfileVerifier: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
    fun f () = 
      let val l = newList n
	  val l = ListMetaAdaptive.fromListEq (InElt.eq, l)
          val l = ListInType.Converter.convert l
      in NonAdaptiveUnAutoApp.run l;("",true) end
  in
    profileN "profileVerifier" (f,rounds)
  end

fun profileVerifierCPS (n,rounds) = 
  let
    val _ = print ("ProfileVerifier: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
    fun f () = 
      let val l = newList n
	  val l = ListMetaAdaptive.fromListEq (InElt.eq, l)
          val l = ListInType.Converter.convert l
      in NonAdaptiveAutoApp.run l;("",true) end
  in
    profileN "profileVerifier" (f,rounds)
  end



fun profileNonAdaptiveAuto (n,rounds) = 
  let
    val _ = print ("ProfileNonAdaptiveAuto: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
    fun f () = 
      let val l = newList n
	  val l = ListMetaAdaptive.fromListEq (InElt.eq, l)
          val l = ListInType.Converter.convert l
      in NonAdaptiveAutoApp.run l;("",true) end
  in
    profileN "profileNonAdaptiveAuto" (f,rounds)
  end
fun profileNonAdaptiveUnAuto (n,rounds) = 
  let
    val _ = print ("ProfileNonAdaptiveUnAuto: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
    fun f () = 
      let val l = newList n
	  val l = ListMetaAdaptive.fromListEq (InElt.eq, l)
          val l = ListInType.Converter.convert l
      in NonAdaptiveUnAutoApp.run l;("",true) end
  in
    profileN "profileNonAdaptiveUnAuto" (f,rounds)
  end

fun profileInitialRun(n,rounds) =
  let
    val _ = print ("ProfileInitialRun: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
   testInitialRun' (n,rounds,PROFILE_TIME)
  end

fun profileMultiInsert(n,rounds) =
  let
    val _ = print ("ProfileMultiInsert: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testMultiInsert' (n,rounds,PROFILE_TIME)
  end

fun profileMultiInsertBatch(n,k, rounds) =
  let
    val _ = print ("ProfileMultiInsertBatch: n = " ^ Int.toString n ^ 
		   " k = " ^ Int.toString k ^ 
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testMultiInsertBatch' (n,k,rounds,PROFILE_TIME)
  end

fun profileMultiInsertBatch(n,k,rounds) =
  let
    val _ = print ("ProfileMultiInsert: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testMultiInsert' (n,rounds,PROFILE_TIME)
  end


fun profileMultiDeleteInsert(n,rounds) =
  let
    val _ = print ("ProfileMultiDeleteInsert: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testMultiDeleteInsert' (n,rounds,PROFILE_TIME)
  end

fun profileAllInsertDelete (n,rounds) =
  let
    val _ = print ("ProfileAllInsertDelete:  = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testAllInsertDelete' (n,rounds,PROFILE_TIME)
  end

fun profileInsertDeleteAtHead (n,rounds) =
  let
    val _ = print ("ProfileInsertDeleteAtHead:  = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testInsertDelete' (n,rounds,0,PROFILE_TIME)
  end


fun profileAllInsertDeleteK (n,k,rounds) =
  let
    val _ = print ("ProfileAllInsertDeleteK:  = " ^ Int.toString n ^
		   "k = " ^ Int.toString k ^ 
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testAllInsertDeleteK' (n,k,rounds,PROFILE_TIME)
  end

(*
  let
    val _ = print ("ProfileInsertDeleteAtHead:  = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testInsertDelete' (n,rounds,0,PROFILE_TIME)
  end
*)



fun profileBatch (length,k,rounds) = 
  testBatch' (length,k,rounds,PROFILE_TIME)

fun profileLocal (length,k,rounds) = 
  testLocal' (length,k,rounds,PROFILE_TIME)

fun profileChangeAll  f (length,rounds) =
  testChangeAll' f (length,rounds,PROFILE_TIME)

fun profileChangeAllK  f (length,k,rounds) =
  testChangeAllK' f (length,k,rounds,PROFILE_TIME)



(**********************************************************************
 ** Space Profiling
 **********************************************************************)
fun sprofileInitialRun(n,rounds) =
  let
    val _ = print ("ProfileInitialRun: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testInitialRun' (n,rounds,PROFILE_SPACE)
  end

fun profileSpace(n,rounds) = sprofileInitialRun(n,rounds)

fun sprofileMultiInsert(n,rounds) =
  let
    val _ = print ("ProfileMultiInsert: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testMultiInsert' (n,rounds,PROFILE_SPACE)
  end


fun sprofileMultiDeleteInsert(n,rounds) =
  let
    val _ = print ("ProfileMultiDeleteInsert: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testMultiDeleteInsert' (n,rounds,PROFILE_SPACE)
  end

fun sprofileAllInsertDelete (n,rounds) =
  let
    val _ = print ("ProfileAllInsertDelete:  = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
    testAllInsertDelete' (n,rounds,PROFILE_SPACE)
  end

fun sprofileChangeAll  f (length,rounds) =
  testChangeAll' f (length,rounds,PROFILE_SPACE)


(**********************************************************************
 ** Mutators
 **********************************************************************)

fun initialRunMutator {size:int} () =
   let
      val l = newList size
      val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l)
      fun next () = NONE
   in
      (i0, (), next)
   end
val initialRunMutator : {size:int} -> unit mutator = initialRunMutator

datatype ins_del = INS | DEL | STOP
fun insertDeleteMutator {size,pos} () =
   let
      val l = newList size
      val h = hd l
      val t = tl l
      val i0 = ListMetaAdaptive.fromListEq (InElt.eq, t)
      val p = ListMetaAdaptive.drop (i0, pos)
      fun next st =
         case st of
            INS =>
               let
                  val _ = ListMetaAdaptive.insertAtEq (InElt.eq, p, h)
               in
                  SOME DEL
               end
          | DEL =>
               let
                  val _ = ListMetaAdaptive.deleteAt p
               in
                  SOME STOP
               end
          | STOP => NONE
   in
      (i0,INS,next)
   end
val insertDeleteMutator : {size:int, pos:int} -> ins_del mutator = insertDeleteMutator

fun deleteInsertMutator {size,pos} () =
   let
      val l = newList size
      val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l)
      val p = ListMetaAdaptive.drop (i0, pos)
      val next =
         case ListMetaAdaptive.deref p of
            ListAdaptive.NIL => 
               (fn _ => NONE)
          | ListAdaptive.CONS(x,_) => 
               (fn st =>
                case st of
                   DEL =>
                      let
                         val _ = ListMetaAdaptive.deleteAt p
                      in
                         SOME INS
                      end
                 | INS =>
                      let
                         val _ = ListMetaAdaptive.insertAtEq (InElt.eq, p, x)
                      in
                         SOME STOP
                      end
                 | STOP => NONE)
   in
      (i0,DEL,next)
   end
val deleteInsertMutator : {size:int, pos:int} -> ins_del mutator = deleteInsertMutator

fun deleteInsertLastMutator {size} () =
   let
      val l = newList size
      val (i0,p) = ListMetaAdaptive.fromListWithLastEq (InElt.eq, l)
      val next =
         case ListMetaAdaptive.deref p of
            ListAdaptive.NIL => 
               (fn _ => NONE)
          | ListAdaptive.CONS(x,_) => 
               (fn st =>
                case st of
                   DEL =>
                      let
                         val _ = ListMetaAdaptive.deleteAt p
                      in
                         SOME INS
                      end
                 | INS =>
                      let
                         val _ = ListMetaAdaptive.insertAtEq (InElt.eq, p, x)
                      in
                         SOME STOP
                      end
                 | STOP => NONE)
   in
      (i0,DEL,next)
   end
val deleteInsertLastMutator : {size:int} -> ins_del mutator = deleteInsertLastMutator

fun allDeleteInsertMutator {size} () =
   let
      val l = newList size
      val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l)
      fun next (st,p,x) =
         case st of
            DEL => (case ListMetaAdaptive.deref p of
                       ListAdaptive.NIL => NONE
                     | ListAdaptive.CONS (x,t) => 
                          let
                             val _ = ListMetaAdaptive.change (p, ListMetaAdaptive.deref t)
                          in
                             SOME (INS,p,SOME x)
                          end)
          | INS => let
                      val _ = ListMetaAdaptive.insertAtEq (InElt.eq,p,valOf x)
                   in
                      SOME (STOP,p,x)
                   end
          | STOP => (case ListMetaAdaptive.deref p of
                        ListAdaptive.NIL => NONE
                      | ListMetaAdaptive.CONS (x,p) => next (DEL,p,NONE))
   in
      (i0,(DEL,i0,NONE),next)
   end
val allDeleteInsertMutator : {size:int} -> (ins_del * InElt.t ListAdaptive.t * InElt.t option) mutator = allDeleteInsertMutator



fun changeAllMutator {size, f_change} () =
   let
      val l = newList size
      val i0 = ListMetaAdaptive.fromListEq (InElt.eq, l)
      fun next p =
	case ListMetaAdaptive.deref p of
          ListAdaptive.NIL => NONE
        | ListAdaptive.CONS (x,t) => 
          let
	    val _ = ListMetaAdaptive.change (p, ListAdaptive.CONS (f_change x, t))
          in
            SOME t
          end
   in
      (i0,i0,next)
   end

val changeAllMutator 
    : {size:int,f_change:InElt.t->InElt.t} -> InElt.t ListAdaptive.t mutator 
= 
changeAllMutator


end

