(**********************************************************************
 ** TestOrdListInputApp.sml
 **
 ** Copyright 2007 Umut A. Acar, Matthew Fluet, Ruy Ley-Wild
 **
 ** This program is distributed under the terms of the GNU General
 ** Public License as published by the Free Software Foundation.  See
 ** the LICENSE file for the terms and conditions.
 **
 ***************
 ** Description: 
 **********************************************************************)
functor TestOrdListInputApp 
   (
    structure InElt : ELT
    structure InNAElt : ELT
    val convertElt : InElt.t -> InNAElt.t
    structure Generator : GENERATOR 
       where type t = InElt.t list
    structure OutType :
       sig
          structure Adaptive : TYPE
          structure NonAdaptive : TYPE
          structure Checker : CHECKER
             where type a_t = Adaptive.t
             where type na_t = NonAdaptive.t
       end
    structure AdaptiveAutoApp : APPLICATION
       where type it = InElt.t list
       where type ot = OutType.Adaptive.t
    structure NonAdaptiveAutoApp : APPLICATION
       where type it = InNAElt.t list
       where type ot = OutType.NonAdaptive.t
    structure NonAdaptiveUnAutoApp : APPLICATION
       where type it = InNAElt.t list
       where type ot = OutType.NonAdaptive.t) =
struct


structure ListInType = BuildOrdListInOut (structure Elt = InElt
					  structure NAElt = InNAElt
					  val convertElt = convertElt)


structure TestApp =
   TestApp
   (structure InType = ListInType
    structure Generator = 
       struct
          type t = InElt.t list
          fun gen {size, seed} = Generator.gen {size = size, seed = seed}
       end 
    structure OutType = OutType
    structure AdaptiveAutoApp = AdaptiveAutoApp
    structure NonAdaptiveAutoApp = NonAdaptiveAutoApp
    structure NonAdaptiveUnAutoApp = NonAdaptiveUnAutoApp)
open TestApp

(**********************************************************************
 ** Mainline
 **********************************************************************)

val SEED = ref 56720
fun newList n = ((SEED := 1 + !SEED); Generator.gen {size = n, seed = !SEED})

fun testInitialRun' (n,rounds,flag) = 
  let
    val _ = AdaptiveMeta.init ()

    val str = "\ntestInitialRun(" ^ Int.toString(n) ^ ")"
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    val i0 = newList n
    val is0 = iToString(i0)

    fun experiment () = 
    let
      val _ = doIf verbose (fn () => print "\nInitial Run:")
      val r0 = AdaptiveAutoApp.run i0
      val rs0 = oToString(!r0)

      (* Check the answer *)
      val test = checkResult (i0,!r0) "Initial Run"

      val _ = passFail (test)

      val _ = doIf verbose (fn () => pPrint ("\nInput = " ^ is0 ^
 	   		                     "\nResult = " ^ rs0 ^ "\n"))
     in
        ("", test)
     end
     
  in
    runExperiment (flag, "testInitialRun") (experiment,rounds)
  end


(**********************************************************************
 ** testChangeAll
 **********************************************************************)
fun testChangeAll' f (n,rounds,flag) =
  let

    val _ = AdaptiveMeta.init ()

    val str = ("\ntestChangeAll(" ^ Int.toString(n) ^  ")")
    val (verbose,iToString,oToString,pPrint,checkResult,passFail) = initExperiment (flag,str)

    val i0 = newList n
    val ls0 = iToString i0

    fun experiment () = 
      let
        (** Initial run **)
        val _ = doIf verbose (fn () => print ("\nInitial Run"))
        val r0 = AdaptiveAutoApp.run i0
        val rs0 = oToString (!r0)
        val _ = doIf verbose (fn () => print "initial run done\n")

        fun changeAll l =
	  let
          in
    	    case l of
	      nil => true
	    | h::t =>
	      let
	        (* Change the head  *)
                val _ = doIf verbose (fn () => print "Changing\n")
	        val _ = f h
	        val _ = AdaptiveMeta.propagate ()
                val _ = doIf verbose (fn () => print "Propagate Done\n")
	        val ls1 = iToString i0
	        val rs1 = oToString (!r0)
                val test1 = checkResult (i0,(!r0)) "Change"

  	      in
	        test1 andalso (changeAll t)
	      end
        end
      val test = changeAll i0
      val _ = passFail (test)
    in
      ("", test)
    end
  in
    runExperiment (flag, "testChangeAll") (experiment,rounds)
  end



(**********************************************************************
 ** Testing Functions
 **********************************************************************)

fun testInitialRun(n) = testInitialRun' (n,1,(TEST_FLAG))
fun testChangeAll  f length = testChangeAll' f (length,1,(TEST_FLAG))


(**********************************************************************
 ** Profiler Functions
 **********************************************************************)

fun profileVerifierCPS (n,rounds) = 
  let
    val _ = print ("ProfileVerifierCPS: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
    fun f () = 
      let val l = newList n
          val l = ListInType.Converter.convert l
      in NonAdaptiveAutoApp.run l;("",true) end
  in
    profileN "profileVerifierCPS" (f,rounds)
  end

fun profileVerifier (n,rounds) = 
  let
    val _ = print ("ProfileVerifier: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
    fun f () = 
      let val l = newList n
          val l = ListInType.Converter.convert l
      in NonAdaptiveUnAutoApp.run l;("",true) end
  in
    profileN "profileVerifier" (f,rounds)
  end

fun profileVerifierCPS (n,rounds) = 
  let
    val _ = print ("ProfileVerifier: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
    fun f () = 
      let val l = newList n
          val l = ListInType.Converter.convert l
      in NonAdaptiveAutoApp.run l;("",true) end
  in
    profileN "profileVerifier" (f,rounds)
  end

fun profileNonAdaptiveAuto (n,rounds) = 
  let
    val _ = print ("ProfileNonAdaptiveAuto: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
    fun f () = 
      let val l = newList n
          val l = ListInType.Converter.convert l
      in NonAdaptiveAutoApp.run l;("",true) end
  in
    profileN "profileNonAdaptiveAuto" (f,rounds)
  end
fun profileNonAdaptiveUnAuto (n,rounds) = 
  let
    val _ = print ("ProfileNonAdaptiveUnAuto: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
    fun f () = 
      let val l = newList n
          val l = ListInType.Converter.convert l
      in NonAdaptiveUnAutoApp.run l;("",true) end
  in
    profileN "profileNonAdaptiveUnAuto" (f,rounds)
  end

fun profileInitialRun(n,rounds) =
  let
    val _ = print ("ProfileInitialRun: n = " ^ Int.toString n ^
		   " rounds = " ^ Int.toString rounds ^ "\n")
  in
   testInitialRun' (n,rounds,PROFILE_TIME)
  end

fun profileChangeAll  f (length,rounds) =
  testChangeAll' f (length,rounds,PROFILE_TIME)

(**********************************************************************
 ** Mutators
 **********************************************************************)

fun initialRunMutator {size:int} () =
   let
      val l = newList size
      fun next () = NONE
   in
      (l, (), next)
   end
val initialRunMutator : {size:int} -> unit mutator = initialRunMutator

fun changeAllMutator {size, f_change} () =
   let
      val i0 = newList size
      fun next p =
	case p of
          nil => NONE
        | x::t => 
          let
	    val _ = f_change x
          in
            SOME t
          end
   in
      (i0,i0,next)
   end

val changeAllMutator 
    : {size:int,f_change:InElt.t->InElt.t} -> InElt.t list mutator 
= 
changeAllMutator


fun changeOneMutator {size, pos, f_change} () =
   let
      val i0 = newList size
      val i1 = List.drop (i0,(List.length i0) - pos)

      fun next p =
	case p of
          nil => NONE
        | (x::t) => 
          let
	    val _ = f_change x
          in
            SOME nil
          end
   in
      (i0,i1,next)
   end

val changeOneMutator 
    : {size:int, pos:int, f_change:InElt.t->InElt.t} -> InElt.t list mutator 
= 
changeOneMutator


end

